// Copyright 2011 Joe Ferner <joe.ferner@nearinfinity.com>
#include "./result.h"
#include <stdio.h>

node_db_sqlite::Result::Column::Column(sqlite3_stmt* stmt, const int column) {
    this->name = std::string(sqlite3_column_origin_name(stmt, column));

    int type = sqlite3_column_type(stmt, column);
    switch(type) {
        case SQLITE_INTEGER:
            this->binary = false;
            this->type = INT;
            break;
        case SQLITE_FLOAT:
            this->binary = false;
            this->type = NUMBER;
            break;
        case SQLITE_TEXT:
            this->binary = false;
            this->type = STRING;
            break;
        case SQLITE_BLOB:
            this->binary = true;
            this->type = TEXT;
            break;
        default:
        case SQLITE_NULL:
            this->binary = false;
            this->type = STRING;
            break;
    }
}

node_db_sqlite::Result::Column::~Column() {
}

bool node_db_sqlite::Result::Column::isBinary() const {
    return this->binary;
}

std::string node_db_sqlite::Result::Column::getName() const {
    return this->name;
}

node_db::Result::Column::type_t node_db_sqlite::Result::Column::getType() const {
    return this->type;
}

node_db_sqlite::Result::Result(sqlite3* connection, sqlite3_stmt* stmt) throw(node_db::Exception&)
    : columns(NULL),
    totalColumns(0),
    rowNumber(0),
    empty(true),
    connection(connection),
    stmt(stmt) {

    _rowData = NULL;
    _columnLengths = NULL;

    try {
        this->totalColumns = sqlite3_column_count(stmt);
        this->stepResult = sqlite3_step(stmt);

        if(this->stepResult != SQLITE_ROW && this->stepResult != SQLITE_DONE) {
            std::stringstream err;
            err << "Cannot move to first row: " << this->stepResult;
            throw node_db::Exception(err.str());
        }

        if (this->totalColumns > 0) {
            this->empty = false;
            this->columns = new Column*[this->totalColumns];
            this->_columnLengths = new unsigned long[this->totalColumns];
            this->_rowData = new char*[this->totalColumns];
            if (this->columns == NULL) {
                throw node_db::Exception("Could not allocate storage for columns");
            }

            for (uint16_t i = 0; i < this->totalColumns; i++) {
                this->columns[i] = new Column(stmt, i);
                if (this->columns[i] == NULL) {
                    this->totalColumns = i;
                    throw node_db::Exception("Could not allocate storage for column");
                }
            }
        }
    } catch(...) {
        this->free();
        throw;
    }
}

node_db_sqlite::Result::~Result() {
    this->free();
}

void node_db_sqlite::Result::free() throw() {
    this->release();

    if (this->columns != NULL) {
        for (uint16_t i = 0; i < this->totalColumns; i++) {
            delete this->columns[i];
        }
        delete[] this->columns;
        this->columns = NULL;
    }

    if (this->_columnLengths != NULL) {
        delete[] this->_columnLengths;
        this->_columnLengths = NULL;
    }

    if (this->_rowData != NULL) {
        delete[] this->_rowData;
        this->_rowData = NULL;
    }
}

void node_db_sqlite::Result::release() throw() {
    if (this->stmt != NULL) {
        sqlite3_finalize(this->stmt);
        this->stmt = NULL;
    }
}

/*
 * We need to cheat here because calls to sqlite3_column_blob and then a call to sqlite3_step
 * will cause the data returned earlier to become invalid. So here are the cases handled here.
 * - If we have already called sqlite3_step and then next gets called return the current data then
 *   clear the stepResult for next call to either next or hasNext.
 * - If the results are cleared and hasNext gets called update stepResult.
 * - If stepResult are filled in and next gets called just return the data.
 *
 * FIX: node-db should really fix how this process works. at least remove the const from hasNext.
*/
void node_db_sqlite::Result::tryStep() {
    if (this->stepResult == 0) {
        this->stepResult = sqlite3_step(stmt);
    }
}

bool node_db_sqlite::Result::hasNext() const throw() {
    ((Result*)this)->tryStep();
    return (this->stepResult != SQLITE_DONE);
}

char** node_db_sqlite::Result::next() throw(node_db::Exception&) {
    tryStep();
    if (this->stepResult == SQLITE_DONE) {
        return NULL;
    }

    for(int i=0; i<this->totalColumns; i++) {
        this->_rowData[i] = (char*)sqlite3_column_blob(stmt, i);
    }

    this->stepResult = 0;
    this->rowNumber++;

    return this->_rowData;
}

unsigned long* node_db_sqlite::Result::columnLengths() throw(node_db::Exception&) {
    for(int i=0; i<this->totalColumns; i++) {
        this->_columnLengths[i] = sqlite3_column_bytes(this->stmt, i);
    }
    return this->_columnLengths;
}

uint64_t node_db_sqlite::Result::index() const throw(std::out_of_range&) {
    if (this->rowNumber == 0) {
        throw std::out_of_range("Not standing on a row");
    }
    return (this->rowNumber - 1);
}

node_db_sqlite::Result::Column* node_db_sqlite::Result::column(uint16_t i) const throw(std::out_of_range&) {
    if (i >= this->totalColumns) {
        throw std::out_of_range("Wrong column index");
    }
    return this->columns[i];
}

uint64_t node_db_sqlite::Result::insertId() const throw() {
    return sqlite3_last_insert_rowid(connection);
}

uint64_t node_db_sqlite::Result::affectedCount() const throw() {
    return sqlite3_changes(connection);
}

uint16_t node_db_sqlite::Result::warningCount() const throw() {
    return 0;
}

uint16_t node_db_sqlite::Result::columnCount() const throw() {
    return this->totalColumns;
}

uint64_t node_db_sqlite::Result::count() const throw(node_db::Exception&) {
    if (!this->isBuffered()) {
        throw node_db::Exception("Result is not buffered");
    }
    throw node_db::Exception("Not implemented (count)");
}

bool node_db_sqlite::Result::isBuffered() const throw() {
    return false;
}

bool node_db_sqlite::Result::isEmpty() const throw() {
    return this->empty;
}
